#!/usr/bin/python3

import os
import rclpy
from rclpy.node import Node
import yaml
from geometry_msgs.msg import Point, Pose, Quaternion
from visualization_msgs.msg import Marker, MarkerArray
from rbs_utils_interfaces.msg import NamedPose, RelativeNamedPose, AssemblyConfig
from rbs_utils_interfaces.srv import GetGraspPose, GetWorkspace
from tf2_ros import TransformListener, Buffer, TransformException
from rbs_assets_library import get_model_meshes_info
from env_manager.utils import Tf2Broadcaster


class AssemblyConfigService(Node):
    def __init__(self, node_name="assembly_config"):
        super().__init__(node_name)

        # Initialize parameters
        self.declare_parameter("db_folder", "asp-example")
        db_folder = self.get_parameter("db_folder").get_parameter_value().string_value

        # Parse the YAML file
        yaml_file = os.path.join(
            os.getenv("RBS_ASSEMBLY_DIR", ""), db_folder, "rbs_db.yaml"
        )
        self.assembly_config = parse_yaml(yaml_file)

        tf2_broadcaster = Tf2Broadcaster(self)
        for relative_part in self.assembly_config.relative_part:
            tf2_broadcaster.broadcast_tf(
                relative_part.relative_at,
                relative_part.name,
                translation=(
                    relative_part.pose.position.x,
                    relative_part.pose.position.y,
                    relative_part.pose.position.z,
                ),
                rotation=(
                    relative_part.pose.orientation.x,
                    relative_part.pose.orientation.y,
                    relative_part.pose.orientation.z,
                    relative_part.pose.orientation.w
                )
            )


        # Services
        self.workspace_service = self.create_service(
            GetWorkspace, "get_workspace", self.get_workspace_callback
        )
        self.grasp_pose_service = self.create_service(
            GetGraspPose, "get_grasp_pose", self.get_grasp_pose_callback
        )

    def get_workspace_callback(self, request, response):
        self.get_logger().info("Workspace request received")
        response.workspace = self.assembly_config.workspace
        response.ok = True
        return response

    def get_grasp_pose_callback(self, request, response):
        response.grasp_pose = self.assembly_config.grasp_pose
        response.ok = True
        return response

def parse_yaml(yaml_file):
    with open(yaml_file, "r") as file:
        data = yaml.safe_load(file)

    assets_db = data.get("assets_db", "")

    workspace = [Point(**point) for point in data.get("workspace", [])]

    absolute_part = []
    for part in data.get("absolute_part", []):
        position = part["pose"]["position"]
        orientation = part["pose"].get(
            "orientation", {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        )
        pose = Pose(position=Point(**position), orientation=Quaternion(**orientation))
        absolute_part.append(NamedPose(name=part["name"], pose=pose))

    relative_part = []
    for part in data.get("relative_part", []):
        position = part["pose"]["position"]
        orientation = part["pose"].get(
            "orientation", {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        )
        pose = Pose(position=Point(**position), orientation=Quaternion(**orientation))
        relative_part.append(
            RelativeNamedPose(
                name=part["name"], relative_at=part["relative_at"], pose=pose
            )
        )

    grasp_pose = []
    for pose in data.get("grasp_pose", []):
        position = pose["pose"]["position"]
        orientation = pose["pose"].get(
            "orientation", {"x": 0.0, "y": 0.0, "z": 0.0, "w": 1.0}
        )
        pose_obj = Pose(
            position=Point(**position), orientation=Quaternion(**orientation)
        )
        grasp_pose.append(
            RelativeNamedPose(
                name=pose["name"], relative_at=pose["relative_at"], pose=pose_obj
            )
        )

    extra_poses = []

    assembly_config = AssemblyConfig(
        assets_db=assets_db,
        workspace=workspace,
        absolute_part=absolute_part,
        relative_part=relative_part,
        grasp_pose=grasp_pose,
        extra_poses=extra_poses,
    )

    return assembly_config

def main():
    rclpy.init()

    executor = rclpy.executors.SingleThreadedExecutor()
    node = AssemblyConfigService()
    executor.add_node(node)
    try:
        executor.spin()
    except (KeyboardInterrupt, rclpy.executors.ExternalShutdownException):
        node.destroy_node()

if __name__ == "__main__":
    main()
